/*
 *   To compile:
 *
 *   gcc -fPIC -shared -o nvidia_spank.so nvidia_spank.c -lnvidia-ml -L/usr/local/cuda-8.0/targets/x86_64-linux/lib/stubs
 *
 */
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/resource.h>

#include <slurm/spank.h>

#include <nvml.h>
/*
 *  * All spank plugins must define this macro for the
 *   * Slurm plugin loader.
 *    */
SPANK_PLUGIN(nvidia_spank, 1);

nvmlComputeMode_t nvidia_compute = NVML_COMPUTEMODE_DEFAULT;
static int nvidia_compute_found = 0;
#define NCOMPUTE_VAR "SPANK_NVIDIA_COMPUTE"

static nvmlEnableState_t nvidia_accounting = NVML_FEATURE_DISABLED;
static int nvidia_accounting_found = 0;
#define NACCOUNTING_VAR "SPANK_NVIDIA_ACCOUNTING"

static int nvidia_mps = 0;
static int nvidia_mps_found = 0;
#define NMPS_VAR "SPANK_NVIDIA_MPS"

static int _nvidia_compute_opt_process (int val,
                                const char *optarg,
                                int remote);
static int _nvidia_accounting_opt_process (int val,
                                const char *optarg,
                                int remote);

static int _nvidia_mps_opt_process (int val,
                                const char *optarg,
                                int remote);

struct spank_option gpu_compute_opt =
    { "gpu-compute", "[default|exclusive-thread|exclusive-process|exclusive]",
      "Select NVidia GPU compute mode, it can be set to default, exclusive-thread and exclusive-process (same as exclusive)", 1, 2,
      (spank_opt_cb_f) _nvidia_compute_opt_process
    };

struct spank_option gpu_accounting_opt =
    { "gpu-accounting", NULL,
      "Turn on NVidia GPU accounting", 0, NVML_FEATURE_ENABLED,
      (spank_opt_cb_f) _nvidia_accounting_opt_process
    };

struct spank_option gpu_mps_opt =
    { "gpu-mps", NULL,
      "Start MPS server, it will automaticall switch to exclusive thread mode", 0, 1,
      (spank_opt_cb_f) _nvidia_mps_opt_process
    };

int slurm_spank_init (spank_t sp, int ac, char **av)
{
    int sc = spank_context();
    if (sc == S_CTX_ALLOCATOR || sc == S_CTX_REMOTE || sc == S_CTX_LOCAL) {
      spank_option_register (sp, &gpu_compute_opt);
      spank_option_register (sp, &gpu_accounting_opt);
      spank_option_register (sp, &gpu_mps_opt);
    }
    return (0);
}

const char * convertToComputeModeString(nvmlComputeMode_t mode)
{
    switch (mode)
    {
        case NVML_COMPUTEMODE_DEFAULT:
            return "default";
        case NVML_COMPUTEMODE_EXCLUSIVE_THREAD:
            return "exclusive-thread";
        case NVML_COMPUTEMODE_PROHIBITED:
            return "prohibited";
        case NVML_COMPUTEMODE_EXCLUSIVE_PROCESS:
            return "exclusive-process";
        default:
            return "unknown";
    }
}

int convertToComputeMode(const char *set, nvmlComputeMode_t *whichone) {
   if (!strcmp(set, "default")) {
       *whichone=NVML_COMPUTEMODE_DEFAULT;
   } else if (!strcmp(set, "exclusive-thread")) {
       *whichone=NVML_COMPUTEMODE_EXCLUSIVE_THREAD;
   } else if (!strcmp(set, "prohibited")) {
       *whichone=NVML_COMPUTEMODE_PROHIBITED;
   } else if ((!strcmp(set, "exclusive-process"))||(!strcmp(set, "exclusive"))) {
       *whichone=NVML_COMPUTEMODE_EXCLUSIVE_PROCESS;
   } else {
       slurm_error("Unknown option %s for --gpu-compute",set);
       return -1;
   }
   return 0;
}

int get_set_nvml_options() {
    nvmlReturn_t result;
    unsigned int device_count, i, toreturn=0;

    // First initialize NVML library
    result = nvmlInit();
    if (NVML_SUCCESS != result)
    {
        slurm_error("Failed to initialize NVML: %s", nvmlErrorString(result));

        return 1;
    }

    result = nvmlDeviceGetCount(&device_count);
    if (NVML_SUCCESS != result)
    {
        slurm_error("Failed to query device count: %s", nvmlErrorString(result));
        toreturn=1;
        goto Goback;
    }
    slurm_info("Found %d device%s\n\n", device_count, device_count != 1 ? "s" : "");

    for (i = 0; i < device_count; i++) {
       nvmlDevice_t device;
       nvmlComputeMode_t current_nvidia_compute;
       nvmlEnableState_t current_nvidia_accounting;

       result = nvmlDeviceGetHandleByIndex(i, &device);
       if (NVML_SUCCESS != result) {
           slurm_error("Failed to get handle for device %i: %s", i, nvmlErrorString(result));
           toreturn=1;
           goto Goback;
       }

       // First deal with compute mode
       result = nvmlDeviceGetComputeMode(device, &current_nvidia_compute);
       if (NVML_ERROR_NOT_SUPPORTED == result)
           slurm_error("This is not CUDA capable device");
       else if (NVML_SUCCESS != result)
       {
           slurm_error("Failed to get compute mode for device %i: %s", i, nvmlErrorString(result));
           toreturn=1;
           goto Goback;
       } else {
           if (current_nvidia_compute == nvidia_compute) {
               slurm_info("Current compute mode is already set to desired compute mode %s",convertToComputeModeString(current_nvidia_compute));
           } else {
               slurm_info("Changing compute mode from %s to %s on device %d", convertToComputeModeString(current_nvidia_compute), convertToComputeModeString(nvidia_compute),i);
               result = nvmlDeviceSetComputeMode(device, nvidia_compute);
               if (NVML_ERROR_NO_PERMISSION == result) {
                   slurm_error("Failed to change compute mode on device %d, have to be root?",i);
               } else if (NVML_ERROR_NOT_SUPPORTED == result) {
                   slurm_error("Failed to change compute mode to %s on device %d, not supported for this device type.",convertToComputeModeString(nvidia_compute),i);
               } else if (NVML_SUCCESS != result) {
                   slurm_error("Failed to set compute mode to %s on device %d due to %s",convertToComputeModeString(nvidia_compute),i,nvmlErrorString(result));
               } else {
                   slurm_info("Succsefully changed compute mode to %s on device %d.",convertToComputeModeString(nvidia_compute),i);
               }
           }
       }

       // And now with GPU accounting
       result = nvmlDeviceGetAccountingMode(device,&current_nvidia_accounting);
       if (NVML_ERROR_NOT_SUPPORTED == result)
           slurm_error("This is not CUDA capable device");
       else if (NVML_SUCCESS != result)
       {
           slurm_error("Failed to get current accounting mode for device %i: %s", i, nvmlErrorString(result));
           toreturn=1;
           goto Goback;
       } else {
           if (current_nvidia_accounting == nvidia_accounting) {
               slurm_info("Current accounting mode is already set to desired mode %d",current_nvidia_accounting);
           } else {
               slurm_info("Changing accounting mode from %d to %d on device %d", current_nvidia_accounting, nvidia_accounting, i);
               result = nvmlDeviceSetAccountingMode(device, nvidia_accounting);
               if (NVML_ERROR_NO_PERMISSION == result) {
                   slurm_error("Failed to change accounting mode on device %d, have to be root?",i);
               } else if (NVML_ERROR_NOT_SUPPORTED == result) {
                   slurm_error("Failed to change accounting mode to %d on device %d, not supported for this device type.",nvidia_accounting,i);
               } else if (NVML_SUCCESS != result) {
                   slurm_error("Failed to set accounting mode to %d on device %d due to %s",nvidia_accounting,i,nvmlErrorString(result));
               } else {
                   slurm_info("Succsefully changed accounting mode to %d on device %d.",nvidia_accounting,i);
               }
           }
       }
    }

Goback:
    result = nvmlShutdown();
    if (NVML_SUCCESS != result)
        slurm_error("Failed to shutdown NVML: %s", nvmlErrorString(result));

    return toreturn;;

}

/* We want srun to inherit sbatch/salloc options so check here if we didn't ge any
 * and inherit them otherwise
 *
 */
int slurm_spank_init_post_opt(spank_t sp, int ac, char **av)
{
    if(spank_context()==S_CTX_ALLOCATOR) {
        if (nvidia_compute_found==1)
	    setenv(NCOMPUTE_VAR, convertToComputeModeString(nvidia_compute), 2);
        if (nvidia_accounting_found==1)
	    setenv(NACCOUNTING_VAR, "1", 2);
        if (nvidia_mps_found==1)
            setenv(NMPS_VAR, "1", 2);
    }
}

int slurm_spank_task_init_privileged(spank_t sp, int ac, char **av)
{
    char val[1024]={0};
    int i;

    if(!nvidia_compute_found && (!spank_getenv(sp,NCOMPUTE_VAR,val,512))) {
        slurm_info("Do not have compute option, will use %s from environment.",val);
        convertToComputeMode(val, &nvidia_compute);
    }
    if(!nvidia_accounting_found && (!spank_getenv(sp,NACCOUNTING_VAR,val,512))) {
        slurm_info("Do not have accounting option, will use the one from environment.");
        nvidia_accounting=NVML_FEATURE_ENABLED;
    }

    slurm_info("Nvidia spank plugin nvidia_compute=%s, nvidia_accounting=%d, mps=%d",
                  convertToComputeModeString(nvidia_compute), nvidia_accounting, nvidia_mps);
    get_set_nvml_options();

    return (0);
}

int slurm_spank_task_init(spank_t sp, int ac, char **av)
{
    char val[1024]={0};
    if(!nvidia_mps_found && (!spank_getenv(sp,NMPS_VAR,val,512))) {
        slurm_info("Do not have mps option, will use the one from environment.");
        nvidia_mps=1;
    }

    if (nvidia_mps) {
        // Try to check if running, assume failure means that it is not
        if(WEXITSTATUS(system("echo get_server_list | /bin/nvidia-cuda-mps-control >/dev/null 2>&1"))==1) {
            slurm_info("Starting MPS");
            setenv("CUDA_DEVICE_MAX_CONNECTIONS","8",0);
            setenv("CUDA_MPS_LOG_DIRECTORY","/tmp",0);
            system("/bin/nvidia-cuda-mps-control -d");
        }
    }
}

static int _nvidia_compute_opt_process (int val,
                                const char *optarg,
                                int remote) {
    if (nvidia_mps) {
        slurm_error("MPS option is on, ignoring compute option %s",optarg);
    } else if(optarg) {
        nvidia_compute_found = 1;
        return convertToComputeMode(optarg, &nvidia_compute);
    }
    return (0);
}

static int _nvidia_accounting_opt_process (int val,
                                const char *optarg,
                                int remote) {
    nvidia_accounting = NVML_FEATURE_ENABLED;
    nvidia_accounting_found = 1;
    return (0);
}

static int _nvidia_mps_opt_process (int val,
                                const char *optarg,
                                int remote) {
    nvidia_mps = 1;
    nvidia_mps_found = 1;
    // also force exclusive mode for compute
    nvidia_compute_found = 1;
    nvidia_compute = NVML_COMPUTEMODE_EXCLUSIVE_PROCESS;
    return (0);
}
